/**
 * Author: Moritz Hanke <hanke@dkrz.de>
 *
 * This program demonstrates a bug that occurs when using cray-mpich.
 * The following MPI implementations have been tested and exhibit this bug:
 * cray-mpich/7.4.3
 * cray-mpich/7.7.4
 *
 * this problem does not occur with the following MPI implementations:
 * intelmpi/2018.4.274
 *
 * Build and call the program (cray) by:
 *   cc test.c
 *   aprun -n 1 a.out
 *
 * In case of a bug the program delivers an output similar to the following:
 * ERROR: buffer[buffer_size-1] != -1
 *
 * Description:
 * It seem like cray-mpich has a problem when packing doubles using
 * MPI_Pack_external to an unaligned address. The packing overwrites data after
 * the position indicated by position. This can cause memory problem, when bytes
 * after the end of the originial buffer are written.
 */

#include <stdlib.h>
#include <stdio.h>
#include <mpi.h>

static const char datarep[] = "external32";

//taken from http://beige.ucs.indiana.edu/I590/node85.html
static void mpi_err_handler(int error_code, MPI_Comm comm) {

   if (error_code != MPI_SUCCESS) {

      int rank;

      MPI_Comm_rank(comm, &rank);

      char error_string[1024];
      int length_of_error_string, error_class;

      MPI_Error_class(error_code, &error_class);
      MPI_Error_string(error_class, error_string, &length_of_error_string);
      fprintf(stderr, "%3d: %s\n", rank, error_string);
      MPI_Error_string(error_code, error_string, &length_of_error_string);
      fprintf(stderr, "%3d: %s\n", rank, error_string);
      MPI_Abort(comm, error_code);
   }
}

int main (void) {

  MPI_Init(NULL, NULL);

  MPI_Aint pack_size;
  mpi_err_handler(MPI_Pack_external_size(datarep, 1, MPI_DOUBLE, &pack_size),
                  MPI_COMM_WORLD);
  if (pack_size != sizeof(double)) {
    fputs("ERROR: pack size of double != sizeof(double)", stderr);
    MPI_Abort(MPI_COMM_WORLD, 1);
  }
  mpi_err_handler(MPI_Pack_external_size(datarep, 1, MPI_CHAR, &pack_size),
                  MPI_COMM_WORLD);
  if (pack_size != 1) {
    fputs("ERROR: pack size of char != 1", stderr);
    MPI_Abort(MPI_COMM_WORLD, 1);
  }

  MPI_Aint buffer_size = sizeof(double) + 2;
  MPI_Aint pack_buffer_size = buffer_size - 1;
  char * buffer = malloc(buffer_size);

  buffer[buffer_size-1] = -1;

  char char_buffer = 1;
  double double_buffer = 1.0;
  MPI_Aint position = 0;
  mpi_err_handler(MPI_Pack_external(datarep, (const void*)&char_buffer, 1,
                                    MPI_CHAR, (void*)buffer, pack_buffer_size,
                                    &position), MPI_COMM_WORLD);
  mpi_err_handler(MPI_Pack_external(datarep, (const void*)&double_buffer, 1,
                                    MPI_DOUBLE, (void*)buffer, pack_buffer_size,
                                    &position), MPI_COMM_WORLD);

  if (position != sizeof(double) + 1) {
    fputs("ERROR: position != sizeof(double) + 1\n", stderr);
    MPI_Abort(MPI_COMM_WORLD, 1);
  }

  // the last byte in the buffer should not have been touched by
  // MPI_Pack_external
  if (buffer[buffer_size-1] != -1) {
    fputs("ERROR: buffer[buffer_size-1] != -1\n", stderr);
    MPI_Abort(MPI_COMM_WORLD, 1);
  }

  free(buffer);

  MPI_Finalize();

  return EXIT_SUCCESS;
}

