/**
 * @file fields.c
 *
 * @copyright Copyright  (C)  2013 Moritz Hanke <hanke@dkrz.de>
 *                                 Rene Redler <rene.redler@mpimet.mpg.de>
 *
 * @version 1.0
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler <rene.redler@mpimet.mpg.de>
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler <rene.redler@mpimet.mpg.de>
 * URL: https://dkrz-sw.gitlab-pages.dkrz.de/yac/
 *
 * This file is part of YAC.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are  permitted provided that the following conditions are
 * met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the DKRZ GmbH nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdlib.h>
#include <string.h>
#include "event.h"
#include "utils.h"
#include "component.h"
#include "fields.h"
#include "interpolation.h"

struct coupling_field {

  // general field related information

  char * name;

  int local_comp_id;
  int global_comp_id;

  struct yac_basic_grid * grid;

  struct interp_field * interp_fields;
  size_t num_interp_fields;
  size_t collection_size; //!< number of vertical levels or bundles

  // data exchange

  enum yac_field_exchange_type exchange_type;

  union {
    struct {
      struct {
        int event_id;
        struct interpolation * interpolation;
        int time_accumulation_count;
        double *** send_field_acc;
      } * puts;
      unsigned num_puts;
    } put;
    struct {
      int event_id;
      struct interpolation * interpolation;
    } get;
  } exchange_data;
};

struct coupling_field *
yac_coupling_field_new(char const * field_name, const int local_comp_id,
                       struct yac_basic_grid * grid,
                       struct interp_field * interp_fields,
                       unsigned num_interp_fields) {

  struct coupling_field * cpl_field = xmalloc(1 * sizeof(*cpl_field));
  struct component * component = yac_unique_id_to_pointer(local_comp_id);

  if ( component != NULL ) 
    cpl_field->global_comp_id  = yac_component_get_global_id(component);
  else
    cpl_field->global_comp_id = -1;

  cpl_field->name              = xmalloc((strlen(field_name) + 1));
  strcpy(cpl_field->name, field_name);
  cpl_field->local_comp_id     = local_comp_id;
  cpl_field->grid              = grid;
  cpl_field->num_interp_fields = (size_t)num_interp_fields;
  cpl_field->collection_size   = 0;
  cpl_field->exchange_type     = NOTHING;

  if (num_interp_fields != 0) {

    if (interp_fields == NULL)
      yac_internal_abort_message(
        "ERROR: num_interp_fields > 0 but not interpolation field provided",
        __FILE__, __LINE__);

    cpl_field->interp_fields =
      xmalloc((size_t)num_interp_fields * sizeof(*(cpl_field->interp_fields)));
    memcpy(cpl_field->interp_fields, interp_fields,
           (size_t)num_interp_fields * sizeof(*(cpl_field->interp_fields)));

  } else  cpl_field->interp_fields = NULL;

  return cpl_field;
}

unsigned yac_get_coupling_field_collection_size(struct coupling_field * field) {

   return field->collection_size;
}

void yac_set_coupling_field_collection_size(struct coupling_field * field, unsigned collection_size ) {

   field->collection_size = collection_size;
}

size_t yac_coupling_field_get_num_interp_fields(struct coupling_field * field) {

  return field->num_interp_fields;
}

const char * yac_get_coupling_field_name(struct coupling_field * field) {

  return field->name;
}

struct yac_basic_grid * yac_coupling_field_get_basic_grid(
  struct coupling_field * field) {

  return field->grid;
}

int yac_get_coupling_field_global_comp_id(struct coupling_field * field) {

  return field->global_comp_id;
}

struct component * yac_get_coupling_field_component(struct coupling_field * field) {

  return (struct component *)yac_unique_id_to_pointer(field->local_comp_id);
}

enum yac_field_exchange_type
yac_get_coupling_field_exchange_type(struct coupling_field * field) {

  return field->exchange_type;
}

static size_t get_interp_field_size(
  struct interp_field field, struct yac_basic_grid * grid) {

  return yac_basic_grid_get_data_size(grid, field.location);
}

size_t yac_coupling_field_get_data_size(
  struct coupling_field * field, enum yac_location location) {

  return yac_basic_grid_get_data_size(field->grid, location);
}

double ***
yac_get_coupling_field_put_op_send_field_acc(struct coupling_field * field,
                                             unsigned put_idx) {

  if (field->exchange_type != SOURCE)
    yac_internal_abort_message("ERROR: wrong field exchange type", __FILE__, __LINE__);

  if (put_idx >= field->exchange_data.put.num_puts)
    yac_internal_abort_message("ERROR: put_idx is invalid", __FILE__, __LINE__);

  if (field->exchange_data.put.puts[put_idx].send_field_acc != NULL)
    return field->exchange_data.put.puts[put_idx].send_field_acc;

  size_t num_interp_fields = field->num_interp_fields;
  size_t collection_size = field->collection_size;

  /* Allocate memory for the accumulation */

  double *** send_field_acc =
    xmalloc(collection_size * sizeof(*send_field_acc));

  for (size_t h = 0; h < collection_size; ++h) {

    send_field_acc[h] = xmalloc(num_interp_fields * sizeof(**send_field_acc));

    for (size_t j = 0; j < num_interp_fields; j++)
      send_field_acc[h][j] =
        xmalloc(
          get_interp_field_size(
            field->interp_fields[j], field->grid) * sizeof(***send_field_acc));
  }

  field->exchange_data.put.puts[put_idx].send_field_acc = send_field_acc;

  /* Initialise memory for the accumulation */

  yac_init_coupling_field_put_op_send_field_acc(field, put_idx, 0.0);
  
  return send_field_acc;
}

void
yac_init_coupling_field_put_op_send_field_acc(struct coupling_field * field,
                                              unsigned put_idx,
                                              double init_value) {

  if (field->exchange_type != SOURCE)
    yac_internal_abort_message("ERROR: wrong field exchange type", __FILE__, __LINE__);

  if (put_idx >= field->exchange_data.put.num_puts)
    yac_internal_abort_message("ERROR: put_idx is invalid", __FILE__, __LINE__);

  double *** send_field_acc =
    (field->exchange_data.put.puts[put_idx].send_field_acc == NULL)?
    yac_get_coupling_field_put_op_send_field_acc(field, put_idx):
    field->exchange_data.put.puts[put_idx].send_field_acc;

  size_t num_interp_fields = field->num_interp_fields;
  size_t collection_size = field->collection_size;

  for (size_t h = 0; h < collection_size; ++h) {
    for (size_t j = 0; j < num_interp_fields; j++) {
      size_t num_points =
        get_interp_field_size(field->interp_fields[j], field->grid);
      for (size_t m = 0; m < num_points; m++ ) {
        send_field_acc[h][j][m] = init_value;
      }
    }
  }
}

int yac_get_coupling_field_put_op_event_id(struct coupling_field * field,
                                           unsigned put_idx) {

  if (field->exchange_type != SOURCE)
    yac_internal_abort_message("ERROR: wrong field exchange type", __FILE__, __LINE__);

  if (put_idx >= field->exchange_data.put.num_puts)
    yac_internal_abort_message("ERROR: put_idx is invalid", __FILE__, __LINE__);

  return field->exchange_data.put.puts[put_idx].event_id;
}

struct interpolation *
yac_get_coupling_field_put_op_interpolation(struct coupling_field * field,
                                            unsigned put_idx) {

  if (field->exchange_type != SOURCE)
    yac_internal_abort_message("ERROR: wrong field exchange type", __FILE__, __LINE__);

  if (put_idx >= field->exchange_data.put.num_puts)
    yac_internal_abort_message("ERROR: put_idx is invalid", __FILE__, __LINE__);

  return field->exchange_data.put.puts[put_idx].interpolation;
}

int
yac_get_coupling_field_put_op_time_accumulation_count(struct coupling_field * field,
                                                      unsigned put_idx) {

  if (field->exchange_type != SOURCE)
    yac_internal_abort_message("ERROR: wrong field exchange type", __FILE__, __LINE__);

  if (put_idx >= field->exchange_data.put.num_puts)
    yac_internal_abort_message("ERROR: put_idx is invalid", __FILE__, __LINE__);

  return field->exchange_data.put.puts[put_idx].time_accumulation_count;
}

void
yac_set_coupling_field_put_op_time_accumulation_count(struct coupling_field * field,
                                                      unsigned put_idx, int count) {

  if (field->exchange_type != SOURCE)
    yac_internal_abort_message("ERROR: wrong field exchange type", __FILE__, __LINE__);

  if (put_idx >= field->exchange_data.put.num_puts)
    yac_internal_abort_message("ERROR: put_idx is invalid", __FILE__, __LINE__);

  field->exchange_data.put.puts[put_idx].time_accumulation_count = count;
}

unsigned yac_get_coupling_field_num_puts(struct coupling_field * field) {

  if (field->exchange_type != SOURCE)
    yac_internal_abort_message("ERROR: wrong field exchange type", __FILE__, __LINE__);

  return field->exchange_data.put.num_puts;
}

int yac_get_coupling_field_get_op_event_id(struct coupling_field * field) {

  if (field->exchange_type != TARGET)
    yac_internal_abort_message("ERROR: wrong field exchange type", __FILE__, __LINE__);

  return field->exchange_data.get.event_id;
}

struct interpolation *
yac_get_coupling_field_get_op_interpolation(struct coupling_field * field) {

  if (field->exchange_type != TARGET)
    yac_internal_abort_message("ERROR: wrong field exchange type", __FILE__, __LINE__);

  return field->exchange_data.get.interpolation;
}

void yac_set_coupling_field_put_op(struct coupling_field * field, int event_id,
                                   struct interpolation * interpolation) {

   if (field->exchange_type == TARGET)
      yac_internal_abort_message("ERROR: a get operation has already been set\n",
                                 __FILE__, __LINE__);

   if (field->exchange_type == NOTHING) {
      field->exchange_data.put.puts = NULL;
      field->exchange_data.put.num_puts = 0;
      field->exchange_type = SOURCE;
   }

   unsigned num_puts = field->exchange_data.put.num_puts;
   field->exchange_data.put.num_puts++;

   field->exchange_data.put.puts =
      xrealloc(field->exchange_data.put.puts, (num_puts + 1) *
               sizeof(*(field->exchange_data.put.puts)));

   field->exchange_data.put.puts[num_puts].event_id = event_id;
   field->exchange_data.put.puts[num_puts].interpolation = interpolation;
   field->exchange_data.put.puts[num_puts].time_accumulation_count = 0;
   field->exchange_data.put.puts[num_puts].send_field_acc = NULL;
}

void yac_set_coupling_field_get_op(struct coupling_field * field, int event_id,
                                   struct interpolation * interpolation) {

   if (field->exchange_type != NOTHING)
      yac_internal_abort_message("ERROR: a put or get operation has already "
                                 "been set\n", __FILE__, __LINE__);

   field->exchange_type = TARGET;
   field->exchange_data.get.event_id = event_id;
   field->exchange_data.get.interpolation = interpolation;
}

struct interp_field const * yac_coupling_field_get_interp_fields(
  struct coupling_field * cpl_field) {

  return cpl_field->interp_fields;
}

size_t yac_coupling_field_num_interp_fields(
  struct coupling_field * cpl_field) {

  return cpl_field->num_interp_fields;
}

void yac_coupling_field_delete(struct coupling_field * cpl_field) {

  size_t collection_size = cpl_field->collection_size;
  size_t num_interp_fields = cpl_field->num_interp_fields;

  if (cpl_field->exchange_type == SOURCE) {

    for (unsigned put_idx = 0; put_idx < cpl_field->exchange_data.put.num_puts;
         ++put_idx) {

      int event_id = cpl_field->exchange_data.put.puts[put_idx].event_id;
      struct interpolation * interpolation =
        cpl_field->exchange_data.put.puts[put_idx].interpolation;
      double ***send_field_acc =
        cpl_field->exchange_data.put.puts[put_idx].send_field_acc;

      if (send_field_acc != NULL) {
        for (size_t h = 0; h < collection_size; ++h) {
          for (size_t j = 0; j < num_interp_fields; ++j)
            free(send_field_acc[h][j]);
          free(send_field_acc[h]);
        }
        free(send_field_acc);
      }

      yac_event_free(event_id);
      yac_interpolation_delete(interpolation);
    }

    free(cpl_field->exchange_data.put.puts);

  } else if (cpl_field->exchange_type == TARGET) {

    yac_event_free(cpl_field->exchange_data.get.event_id);
    yac_interpolation_delete(cpl_field->exchange_data.get.interpolation);
  }

  free(cpl_field->name);
  free(cpl_field->interp_fields);
  free(cpl_field);
}
