/**
 * @file grid.c
 *
 * @copyright Copyright  (C)  2013 Moritz Hanke <hanke@dkrz.de>
 *                                 Rene Redler <rene.redler@mpimet.mpg.de>
 *                                 Thomas Jahns <jahns@dkrz.de>
 *
 * @version 1.0
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler <rene.redler@mpimet.mpg.de>
 *         Thomas Jahns <jahns@dkrz.de>
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler <rene.redler@mpimet.mpg.de>
 *             Thomas Jahns <jahns@dkrz.de>
 * URL: https://dkrz-sw.gitlab-pages.dkrz.de/yac/
 *
 * This file is part of YAC.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are  permitted provided that the following conditions are
 * met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the DKRZ GmbH nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <stdio.h>
#include <string.h>

#include "grid.h"
#include "grid_cell.h"

struct yac_basic_grid {
  char * name;
  struct yac_field_data_set field_data_set;
  struct basic_grid_data data;
};

struct basic_grid_data basic_grid_data_empty = {
  .vertex_coordinates = NULL,
  .cell_ids = NULL,
  .vertex_ids = NULL,
  .edge_ids = NULL,
  .num_cells = 0,
  .num_vertices = 0,
  .num_edges = 0,
  .core_cell_mask = NULL,
  .core_vertex_mask = NULL,
  .core_edge_mask = NULL,
  .num_vertices_per_cell = NULL,
  .num_cells_per_vertex = NULL,
  .cell_to_vertex = NULL,
  .cell_to_vertex_offsets = NULL,
  .cell_to_edge = NULL,
  .cell_to_edge_offsets = NULL,
  .vertex_to_cell = NULL,
  .vertex_to_cell_offsets = NULL,
  .edge_to_vertex = NULL,
  .edge_type = NULL,
  .num_total_cells = 0,
  .num_total_vertices = 0,
  .num_total_edges = 0
};

struct yac_basic_grid * yac_basic_grid_new(
  char const * name, struct basic_grid_data grid_data) {

  struct yac_basic_grid * grid = xmalloc(1 * sizeof(*grid));

  size_t name_size = strlen(name);

  grid->name = xmalloc(name_size + 1);
  strcpy(grid->name, name);
  grid->field_data_set = yac_field_data_set_init();
  grid->data = grid_data;

  return grid;
};

void yac_basic_grid_delete(struct yac_basic_grid * grid) {

  if (grid == NULL) return;
  free(grid->name);
  yac_field_data_set_free(grid->field_data_set);
  yac_basic_grid_data_free(grid->data);
  free(grid);
}

// the order must match the entries in enum  yac_location
static char const * yac_location_str[] = {"CELL",
                                          "CORNER",
                                          "EDGE",
                                          "UNDEFINED"};

enum yac_location yac_str2loc(char const * location) {

  if (!strncmp("CELL", location, YAC_MAX_LOC_STR_LEN))
    return CELL;
  else if (!strncmp("CORNER", location, YAC_MAX_LOC_STR_LEN))
    return CORNER;
  else if (!strncmp("EDGE", location, YAC_MAX_LOC_STR_LEN))
    return EDGE;
  else if (!strncmp("UNDEFINED", location, YAC_MAX_LOC_STR_LEN))
    return LOC_UNDEFINED;
  else
      yac_internal_abort_message("ERROR(yac_str2loc): invalid location\n",
                                 __FILE__, __LINE__);

  // routine will never reach this point
  return LOC_UNDEFINED;
}

char const * yac_loc2str(enum yac_location location) {

  switch (location) {
    case (CELL):
    case (CORNER):
    case (EDGE):
    case (LOC_UNDEFINED):
      break;
    default:
      yac_internal_abort_message("ERROR(yac_loc2str): location must be one of "
                                 "CORNER/EDGE/CELL/LOC_UNDEFINED.", __FILE__,
                                 __LINE__ );
  };

  return yac_location_str[location];
}

enum yac_location yac_get_location(int const location) {

  switch (location) {
    case (CELL):
    case (CORNER):
    case (EDGE):
    case (LOC_UNDEFINED):
      return (enum yac_location) location;
    default:
      yac_internal_abort_message("ERROR: get_location: location must be one "
                                 "of CORNER/EDGE/CELL/LOC_UNDEFINED.", __FILE__,
                                 __LINE__ );
  };

  // routine will never reach this point
  return CORNER;
}

coordinate_pointer yac_basic_grid_get_field_coordinates(
  struct yac_basic_grid * grid, struct interp_field field) {

  coordinate_pointer coords = NULL;

  if (field.coordinates_idx != SIZE_MAX) {

    struct yac_field_data * data =
      yac_basic_grid_get_field_data(grid, field.location);

    if (field.coordinates_idx >= data->coordinates_count)
      yac_internal_abort_message(
        "ERROR(yac_basic_grid_get_field_coordinates): invalid coord_idx",
        __FILE__, __LINE__);

    coords = data->coordinates[field.coordinates_idx];
  }

  // if no field coordinates are defined, but the location is at the corners of
  // of the grid cells, return coordinates of them
  if ((coords == NULL) && (field.location == CORNER))
    coords = grid->data.vertex_coordinates;

  return coords;
}

int const * yac_basic_grid_get_field_mask(
  struct yac_basic_grid * grid, struct interp_field field) {

  if (field.masks_idx == SIZE_MAX) return NULL;

  struct yac_field_data * data =
    yac_basic_grid_get_field_data(grid, field.location);

  if (field.masks_idx >= data->masks_count)
    yac_internal_abort_message(
      "ERROR(yac_basic_grid_get_field_mask): invalid mask_idx",
      __FILE__, __LINE__);

  return data->masks[field.masks_idx];
}

char const * yac_basic_grid_get_name(struct yac_basic_grid * grid) {

  return grid->name;
}

struct basic_grid_data * yac_basic_grid_get_data(struct yac_basic_grid * grid) {

  if (grid != NULL) return &(grid->data);
  else return &basic_grid_data_empty;
}

static struct yac_field_data * get_field_data(
  struct yac_field_data_set * field_data_set, enum yac_location location) {

  switch (location) {
    case (CELL): return &field_data_set->cell;
    case (CORNER): return &field_data_set->vertex;
    case (EDGE): return &field_data_set->edge;
    default:
      yac_internal_abort_message(
        "ERROR(get_field_data): invalid location", __FILE__, __LINE__);
      return NULL;
  };
}

size_t yac_basic_grid_get_data_size(
  struct yac_basic_grid * grid, enum yac_location location) {

  switch (location) {
    case (CELL):
      return grid->data.num_cells;
    case (CORNER):
      return grid->data.num_vertices;
    case (EDGE):
      return grid->data.num_edges;
    default:
      yac_internal_abort_message(
        "ERROR(yac_basic_grid_get_data_size): invalid location",
        __FILE__, __LINE__);
      return 0;
  };
}

size_t yac_basic_grid_add_coordinates_nocpy(
  struct yac_basic_grid * grid,
  enum yac_location location, coordinate_pointer coordinates) {

  return
    yac_field_data_set_add_coordinates_nocpy(
      &grid->field_data_set, location, coordinates);
}

size_t yac_basic_grid_add_coordinates(
  struct yac_basic_grid * grid, enum yac_location location,
  coordinate_pointer coordinates, size_t count) {

  return
    yac_field_data_set_add_coordinates(
      &grid->field_data_set, location, coordinates, count);
}

size_t yac_basic_grid_add_mask_nocpy(
  struct yac_basic_grid * grid,
  enum yac_location location, int const * mask) {

  return
    yac_field_data_set_add_mask_nocpy(
      &grid->field_data_set, location, mask);
}

size_t yac_basic_grid_add_mask(
  struct yac_basic_grid * grid, enum yac_location location,
  int const * mask, size_t count) {

  return
    yac_field_data_set_add_mask(
      &grid->field_data_set, location, mask, count);
}

struct yac_field_data * yac_basic_grid_get_field_data(
  struct yac_basic_grid * grid, enum yac_location location) {

  if (grid != NULL) return get_field_data(&(grid->field_data_set), location);
  else return NULL;
}

void yac_basic_grid_data_free(struct basic_grid_data grid) {

  free(grid.vertex_coordinates);
  free(grid.cell_ids);
  free(grid.vertex_ids);
  free(grid.edge_ids);
  free(grid.core_cell_mask);
  free(grid.core_vertex_mask);
  free(grid.core_edge_mask);
  free(grid.num_vertices_per_cell);
  free(grid.num_cells_per_vertex);
  free(grid.cell_to_vertex);
  free(grid.cell_to_vertex_offsets);
  free(grid.cell_to_edge);
  if (grid.cell_to_vertex_offsets != grid.cell_to_edge_offsets)
    free(grid.cell_to_edge_offsets);
  free(grid.vertex_to_cell);
  free(grid.vertex_to_cell_offsets);
  free(grid.edge_to_vertex);
  free(grid.edge_type);
}

struct yac_field_data yac_field_data_init() {
  struct yac_field_data field_data = {.masks_count = 0,
                                      .masks = NULL,
                                      .coordinates_count = 0,
                                      .coordinates = NULL};
  return field_data;
}

size_t yac_field_data_add_mask_nocpy(
  struct yac_field_data * field_data, int const * mask) {

  size_t masks_idx = field_data->masks_count++;
  field_data->masks =
    xrealloc(field_data->masks,
             field_data->masks_count * sizeof(*(field_data->masks)));
  field_data->masks[masks_idx] = (int*)mask;

  return masks_idx;
}

size_t yac_field_data_add_coordinates_nocpy(
  struct yac_field_data * field_data, coordinate_pointer coordinates) {

  size_t coordinates_idx = field_data->coordinates_count++;
  field_data->coordinates =
    xrealloc(
      field_data->coordinates,
      field_data->coordinates_count * sizeof(*(field_data->coordinates)));
  field_data->coordinates[coordinates_idx] = coordinates;

  return coordinates_idx;
}

struct yac_field_data_set yac_field_data_set_init() {
  struct yac_field_data_set field_data_set =
    {.cell = yac_field_data_init(),
     .vertex = yac_field_data_init(),
     .edge = yac_field_data_init()};
  return field_data_set;
}

size_t yac_field_data_set_add_mask_nocpy(
  struct yac_field_data_set * field_data_set,
  enum yac_location location, int const * mask) {

  return
    yac_field_data_add_mask_nocpy(
      get_field_data(field_data_set, location), mask);
}

size_t yac_field_data_set_add_mask(
  struct yac_field_data_set * field_data_set,
  enum yac_location location, int const * mask,
  size_t count) {

  int * mask_cpy = xmalloc(count * sizeof(*mask_cpy));
  memcpy(mask_cpy, mask, count * sizeof(*mask));

  return
    yac_field_data_set_add_mask_nocpy(field_data_set, location, mask_cpy);
}

size_t yac_field_data_set_add_coordinates_nocpy(
  struct yac_field_data_set * field_data_set,
  enum yac_location location, coordinate_pointer coordinates) {

  return
    yac_field_data_add_coordinates_nocpy(
      get_field_data(field_data_set, location), coordinates);
}

size_t yac_field_data_set_add_coordinates(
  struct yac_field_data_set * field_data_set,
  enum yac_location location, coordinate_pointer coordinates,
  size_t count) {

  coordinate_pointer coordinates_cpy =
    xmalloc(count * sizeof(*coordinates_cpy));
  memcpy(coordinates_cpy, coordinates, count * sizeof(*coordinates));

  return
    yac_field_data_set_add_coordinates_nocpy(
      field_data_set, location, coordinates_cpy);
}

static void yac_field_data_free(struct yac_field_data field_data) {

  for (size_t i = 0; i < field_data.masks_count; ++i)
    free(field_data.masks[i]);
  free(field_data.masks);
  for (size_t i = 0; i < field_data.coordinates_count; ++i)
    free(field_data.coordinates[i]);
  free(field_data.coordinates);
}

void yac_field_data_set_free(struct yac_field_data_set field_data_set) {
  yac_field_data_free(field_data_set.cell);
  yac_field_data_free(field_data_set.vertex);
  yac_field_data_free(field_data_set.edge);
}
