/**
 * @file test_interp_weights_parallel.c
 *
 * @copyright Copyright  (C)  2019 DKRZ, MPI-M
 *
 * @author Moritz Hanke <hanke@dkrz.de>
 *
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 * URL: https://dkrz-sw.gitlab-pages.dkrz.de/yac/
 *
 * This file is part of YAC.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are  permitted provided that the following conditions are
 * met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the DKRZ GmbH nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <stdlib.h>
#include <unistd.h>

#include "tests.h"
#include "test_common.h"
#include "geometry.h"
#include "read_icon_grid.h"
#include "interp_method.h"
#include "interp_method_file.h"
#include "interp_weights.h"
#include "yac_mpi.h"
#include "dist_grid_utils.h"

#if defined(YAC_NETCDF_ENABLED)

#include <mpi.h>
#include <yaxt.h>
#include <netcdf.h>

enum weight_type {
  NONE = 0,
  DIRECT = 1,
  SUM = 2,
  WSUM = 3
};

static void get_basic_grid_data(
  char * filename, size_t * num_cells, size_t * num_vertices,
  size_t * num_edges);
static int check_direct_results(
  struct basic_grid_data * grid, int is_src, struct interp_weights * weights,
  enum interp_weights_reorder_type reorder_type, size_t collection_size,
  double *** src_data, double ** tgt_data);
static int check_results_ref(
  int is_src, struct interp_weights * weights,
  enum interp_weights_reorder_type reorder_type, double * src_data,
  double * tgt_data, double * ref_tgt_data, size_t tgt_size);

static struct basic_grid_data
  generate_dummy_grid_data(
    size_t num_cells, size_t global_num_cells, size_t num_cells_offset);

int main(void) {

  MPI_Init(NULL, NULL);

  xt_initialize(MPI_COMM_WORLD);

  int comm_rank, comm_size;
  yac_mpi_call(MPI_Comm_rank(MPI_COMM_WORLD, &comm_rank), MPI_COMM_WORLD);
  yac_mpi_call(MPI_Comm_size(MPI_COMM_WORLD, &comm_size), MPI_COMM_WORLD);

  if (comm_size != 5) {
    PUT_ERR("This test requires 5 processes\n");
    xt_finalize();
    MPI_Finalize();
    return TEST_EXIT_CODE;
  }

  enum interp_weights_reorder_type reorder_types[] = {
    MAPPING_ON_SRC,
    MAPPING_ON_TGT
  };
  size_t reorder_type_count =
    sizeof(reorder_types) / sizeof(reorder_types[0]);

  { // test with 2 target and 3 source processes
    // * number source and target points is identical
    // * first target process owns all even target cells and the other
    //   all odd target cells
    // * each source process owns one third of the source points
    //   (they are consecutive)
    // * each process generates a part of the weights for a consecutive part
    //   of the target points
    int is_src = comm_rank < 3;

    if (comm_size != 5) {
      PUT_ERR("This test requires 5 processes\n");
      xt_finalize();
      MPI_Finalize();
      return TEST_EXIT_CODE;
    }

    int const src_proc_count = 3;
    int const tgt_proc_count = 2;
    size_t const global_size =
      src_proc_count * tgt_proc_count * (src_proc_count + tgt_proc_count) * 2;
    size_t const local_size = global_size / (src_proc_count + tgt_proc_count);

    double wsum_weights[3*local_size];
    double sum_weights[3*local_size];
    double direct_weights[local_size];
    size_t wsum_num_src_per_tgt[local_size];
    size_t direct_num_src_per_tgt[local_size];
    struct remote_points * default_tgts = xmalloc(1 * sizeof(*default_tgts));
    default_tgts->data = xmalloc(local_size * sizeof(*default_tgts->data));
    struct remote_points none_tgts;
    struct remote_point * wsum_srcs =
      xmalloc(3 * local_size * sizeof(*wsum_srcs));
    struct remote_point * direct_srcs =
      xmalloc(local_size * sizeof(*wsum_srcs));

    default_tgts->count = local_size;
    for (size_t i = 0; i < local_size; ++i) {
      for (size_t j = 0; j < 3; ++j) {
        wsum_weights[3 * i + j] = (double)(j+1);
        sum_weights[3 * i + j] = 1.0;
      }
      direct_weights[i] = 1.0;
      wsum_num_src_per_tgt[i] = 3;
      direct_num_src_per_tgt[i] = 1;

      yac_int tgt_global_id = (yac_int)(comm_rank * local_size + i);
      default_tgts->data[i].global_id = tgt_global_id;
      default_tgts->data[i].data.count = 1;
      default_tgts->data[i].data.data.single.rank =
        (int)(tgt_global_id & 1) + (int)src_proc_count;
      default_tgts->data[i].data.data.single.orig_pos =
        (uint64_t)(tgt_global_id / 2);

      for (size_t j = 0; j < 3; ++j) {
        yac_int src_global_id =
          (tgt_global_id + (yac_int)j)%((yac_int)global_size);
        wsum_srcs[3 * i + j].global_id = src_global_id;
        wsum_srcs[3 * i + j].data.count = 1;
        wsum_srcs[3 * i + j].data.data.single.rank =
          (int)src_global_id % (int)src_proc_count;
        wsum_srcs[3 * i + j].data.data.single.orig_pos =
          (uint64_t)(src_global_id / src_proc_count);
      }

      yac_int src_global_id = tgt_global_id;
      direct_srcs[i].global_id = src_global_id;
      direct_srcs[i].data.count = 1;
      direct_srcs[i].data.data.single.rank =
        (int)src_global_id % (int)src_proc_count;
      direct_srcs[i].data.data.single.orig_pos =
        (uint64_t)(src_global_id / src_proc_count);
    }
    none_tgts.count = 0;
    none_tgts.data = 0;

    double * out_data =
      (is_src)?xmalloc((global_size / src_proc_count) * sizeof(*out_data)):NULL;
    if (is_src)
      for (size_t i = 0; i < global_size / src_proc_count; ++i)
        out_data[i] = (double)((yac_int)comm_rank + i * src_proc_count);
    double * in_data =
      (is_src)?NULL:xmalloc((global_size / tgt_proc_count) * sizeof(*in_data));
    double * ref_in_data =
      (is_src)?NULL:xmalloc((global_size / tgt_proc_count) * sizeof(*ref_in_data));

    struct basic_grid_data grid_data =
      generate_dummy_grid_data(
        global_size / ((is_src)?src_proc_count:tgt_proc_count), global_size,
        (size_t)comm_rank - ((is_src)?0:src_proc_count));

    struct yac_basic_grid * grid =
      yac_basic_grid_new((is_src)?"src_grid":"tgt_grid", grid_data);

    int grid_ids[2] = {2, 3};
    struct dist_grid_pair * grid_pair =
      yac_dist_grid_pair_new(
        grid, grid_ids[is_src], NULL, grid_ids[is_src^1], MPI_COMM_WORLD);

    struct interp_field src_fields[] =
      {{.location = CELL, .coordinates_idx = SIZE_MAX, .masks_idx = SIZE_MAX}};
    size_t num_src_fields = sizeof(src_fields) / sizeof(src_fields[0]);
    struct interp_field tgt_field =
      {.location = CELL, .coordinates_idx = SIZE_MAX, .masks_idx = SIZE_MAX};

    struct interp_grid * interp_grid =
      yac_interp_grid_new(grid_pair, grid_ids[1], grid_ids[0],
                          num_src_fields, src_fields, tgt_field);

    for (size_t i = 0; i < (1 << (2 * (src_proc_count + tgt_proc_count))); ++i) {

      enum weight_type weight_type[src_proc_count + tgt_proc_count];
      int max_weight_type = 0;

      for (size_t j = 0; j < src_proc_count + tgt_proc_count; ++j) {
        weight_type[j] = (enum weight_type)((i >> (3 * j)) & 3);
        if ((int)(weight_type[j]) > max_weight_type)
          max_weight_type = (int)(weight_type[j]);
      }

      if (!is_src) {
        for (size_t j = 0; j < global_size / tgt_proc_count; ++j) {
          yac_int tgt_global_id =
            (yac_int)(comm_rank - (int)src_proc_count) + j * tgt_proc_count;
          // which rank generates the weights
          int weight_rank = (int)(tgt_global_id / (yac_int)local_size);
          double ref_value;
          switch (weight_type[weight_rank]) {
            case(NONE):
            default:
              ref_value = -1.0;
              break;
            case(DIRECT):
              ref_value = (double)tgt_global_id;
              break;
            case(SUM):
              ref_value = (double)(((tgt_global_id + 0) % global_size) +
                                   ((tgt_global_id + 1) % global_size) +
                                   ((tgt_global_id + 2) % global_size));
              break;
            case(WSUM):
              ref_value = (double)(1 * ((tgt_global_id + 0) % global_size) +
                                   2 * ((tgt_global_id + 1) % global_size) +
                                   3 * ((tgt_global_id + 2) % global_size));
              break;
          };
          ref_in_data[j] = ref_value;
          in_data[j] = -1.0;
        }
      }

      struct remote_points * tgts;
      size_t * num_src_per_tgt;
      struct remote_point * srcs;
      double * w;

      switch (weight_type[comm_rank]) {
        case(NONE):
        default:
          tgts = &none_tgts;
          num_src_per_tgt = NULL;
          srcs = NULL;
          w = NULL;
          break;
        case(DIRECT):
          tgts = default_tgts;
          num_src_per_tgt = direct_num_src_per_tgt;
          srcs = direct_srcs;
          w = direct_weights;
          break;
        case(SUM):
          tgts = default_tgts;
          num_src_per_tgt = wsum_num_src_per_tgt;
          srcs = wsum_srcs;
          w = sum_weights;
          break;
        case(WSUM):
          tgts = default_tgts;
          num_src_per_tgt = wsum_num_src_per_tgt;
          srcs = wsum_srcs;
          w = wsum_weights;
          break;
      }

      for (int j = MAX(max_weight_type, DIRECT); j <= (int)WSUM; ++j) {

        struct interp_weights * weights = yac_interp_weights_new(interp_grid);

        switch (j) {
          case(NONE):
          default:
          break;
          case(DIRECT):
            yac_interp_weights_add_direct(weights, tgts, srcs);
            break;
          case(SUM):
            yac_interp_weights_add_sum(weights, tgts, num_src_per_tgt, srcs);
            break;
          case(WSUM):
            yac_interp_weights_add_wsum(
              weights, tgts, num_src_per_tgt, srcs, w);
            break;
        }

        // checking with different reorder types and collection sizes
        for (size_t reorder_type = 0; reorder_type < reorder_type_count;
             ++reorder_type)
          if (check_results_ref(
                is_src, weights, reorder_types[reorder_type],
                out_data, in_data, ref_in_data, global_size / tgt_proc_count))
            PUT_ERR("ERROR(yac_interp_weights_add_*): "
                    "invalid interpolation result");

        yac_interp_weights_delete(weights);
      }
    }
    yac_interp_grid_delete(interp_grid);
    yac_dist_grid_pair_delete(grid_pair);
    yac_basic_grid_delete(grid);
    free(ref_in_data);
    free(in_data);
    free(out_data);
    free(direct_srcs);
    free(wsum_srcs);
    free(default_tgts->data);
    free(default_tgts);
  }

  {
    int is_src = comm_rank < (comm_size / 2);

    MPI_Comm local_grid_comm;
    int local_grid_comm_size;
    yac_mpi_call(
      MPI_Comm_split(
        MPI_COMM_WORLD, is_src, 0, &local_grid_comm), MPI_COMM_WORLD);
    yac_mpi_call(
      MPI_Comm_size(
        local_grid_comm, &local_grid_comm_size), local_grid_comm);

    char * filenames[] ={"icon_grid_R02B02.nc", "icon_grid_R02B03_G.nc"};
    struct basic_grid_data grid_data =
      read_icon_grid_information_parallel2(
        filenames[is_src], local_grid_comm, MIN(local_grid_comm_size, 2));
    struct yac_basic_grid * grid =
      yac_basic_grid_new(filenames[is_src], grid_data);

    int grid_ids[2] = {2, 3};
    struct dist_grid_pair * grid_pair =
      yac_dist_grid_pair_new(
        grid, grid_ids[is_src], NULL, grid_ids[is_src^1], MPI_COMM_WORLD);

    struct interp_field src_fields[] =
      {{.location = CELL, .coordinates_idx = SIZE_MAX, .masks_idx = SIZE_MAX}};
    size_t num_src_fields = sizeof(src_fields) / sizeof(src_fields[0]);
    struct interp_field tgt_field =
      {.location = CELL, .coordinates_idx = SIZE_MAX, .masks_idx = SIZE_MAX};

    struct interp_grid * interp_grid =
      yac_interp_grid_new(grid_pair, grid_ids[1], grid_ids[0],
                          num_src_fields, src_fields, tgt_field);

    struct interp_weights * weights = yac_interp_weights_new(interp_grid);

    size_t num_tgt_cells, num_tgt_vertices, num_tgt_edges;
    get_basic_grid_data(
      filenames[0], &num_tgt_cells, &num_tgt_vertices, &num_tgt_edges);

    {
      size_t count;
      size_t * tgt_points;
      yac_interp_grid_get_tgt_points(interp_grid, &tgt_points, &count);
      yac_int * tgt_global_ids = xmalloc(count * sizeof(*tgt_global_ids));
      size_t * size_t_buffer = xmalloc(2 * count * sizeof(*size_t_buffer));
      size_t * odd_tgt_points = size_t_buffer;
      size_t * even_tgt_points = size_t_buffer + count;
      yac_interp_grid_get_tgt_global_ids(
        interp_grid, tgt_points, count, tgt_global_ids);

      size_t odd_count = 0;
      size_t even_count = 0;
      for (size_t i = 0; i < count; ++i) {
        if (tgt_global_ids[i] & 1) odd_tgt_points[odd_count++] = tgt_points[i];
        else even_tgt_points[even_count++] = tgt_points[i];
      }

      struct remote_points odd_tgts = {
        .data =
          yac_interp_grid_get_tgt_remote_points(
            interp_grid, odd_tgt_points, odd_count),
        .count = odd_count};
      struct remote_points even_tgts = {
        .data =
          yac_interp_grid_get_tgt_remote_points(
            interp_grid, even_tgt_points, even_count),
        .count = even_count};

      yac_interp_weights_add_fixed(weights, &odd_tgts, -1.0);
      yac_interp_weights_add_fixed(weights, &even_tgts, -2.0);

      free(odd_tgts.data);
      free(even_tgts.data);
      free(size_t_buffer);
      free(tgt_global_ids);
      free(tgt_points);
    }

    for (size_t i = 0; i < reorder_type_count; ++i) {

      struct interpolation * interpolation =
        yac_interp_weights_get_interpolation(weights, reorder_types[i], 1);

      double * field_data =
        xmalloc(grid_data.num_cells * sizeof(*field_data));
      if (is_src)
        for (size_t j = 0; j < grid_data.num_cells; ++j)
          field_data[j] = (double)1;
      else
        for (size_t j = 0; j < grid_data.num_cells; ++j)
          field_data[j] = (double)-3;

      double ** src_fields, * tgt_field;
      if (is_src) {
        src_fields = &field_data;
        tgt_field = NULL;
      } else {
        src_fields = NULL;
        tgt_field = field_data;
      }

      yac_interpolation_execute(interpolation, &src_fields, &tgt_field);

      if (!is_src) {
        for (size_t j = 0; j < grid_data.num_cells; ++j) {
          if (grid_data.core_cell_mask[j]) {
            if (grid_data.cell_ids[j] & 1) {
              if (tgt_field[j] != -1.0)
                PUT_ERR("ERROR(yac_interp_weights_add_fixed): "
                        "invalid interpolation result");
            } else {
              if (tgt_field[j] != -2.0)
                PUT_ERR("ERROR(yac_interp_weights_add_fixed): "
                        "invalid interpolation result");
            }
          } else {
            if (tgt_field[j] != -3.0)
              PUT_ERR("ERROR(yac_interp_weights_add_fixed): "
                      "invalid interpolation result");
          }
        }
      }

      free(field_data);
      yac_interpolation_delete(interpolation);
    }

    yac_interp_weights_delete(weights);
    yac_interp_grid_delete(interp_grid);
    yac_dist_grid_pair_delete(grid_pair);
    yac_basic_grid_delete(grid);
    yac_mpi_call(MPI_Comm_free(&local_grid_comm), MPI_COMM_WORLD);
  }

  {
    int is_src = comm_rank < (comm_size / 2);

    MPI_Comm local_grid_comm;
    int local_grid_comm_size;
    yac_mpi_call(
      MPI_Comm_split(
        MPI_COMM_WORLD, is_src, 0, &local_grid_comm), MPI_COMM_WORLD);
    yac_mpi_call(
      MPI_Comm_size(
        local_grid_comm, &local_grid_comm_size), local_grid_comm);

    char * filenames[] ={"icon_grid_R02B02.nc", "icon_grid_R02B02.nc"};
    struct basic_grid_data grid_data =
      read_icon_grid_information_parallel2(
        filenames[is_src], local_grid_comm, MIN(local_grid_comm_size, 2));
    struct yac_basic_grid * grid =
      yac_basic_grid_new(filenames[is_src], grid_data);

    int grid_ids[2] = {2, 3};

    struct dist_grid_pair * grid_pair =
      yac_dist_grid_pair_new(
        grid, grid_ids[is_src], NULL, grid_ids[is_src^1], MPI_COMM_WORLD);

    struct interp_field src_fields[] =
      {{.location = CELL, .coordinates_idx = SIZE_MAX, .masks_idx = SIZE_MAX}};
    size_t num_src_fields = sizeof(src_fields) / sizeof(src_fields[0]);
    struct interp_field tgt_field =
      {.location = CELL, .coordinates_idx = SIZE_MAX, .masks_idx = SIZE_MAX};

    struct interp_grid * interp_grid =
      yac_interp_grid_new(grid_pair, grid_ids[1], grid_ids[0],
                          num_src_fields, src_fields, tgt_field);

    struct interp_weights * weights = yac_interp_weights_new(interp_grid);

    size_t num_tgt_cells, num_tgt_vertices, num_tgt_edges;
    get_basic_grid_data(
      filenames[0], &num_tgt_cells, &num_tgt_vertices, &num_tgt_edges);

    {
      // get all local target points
      size_t count;
      size_t * tgt_points;
      yac_interp_grid_get_tgt_points(interp_grid, &tgt_points, &count);

      // get global ids of local target points
      yac_int * tgt_global_ids = xmalloc(count * sizeof(*tgt_global_ids));
      yac_interp_grid_get_tgt_global_ids(
        interp_grid, tgt_points, count, tgt_global_ids);

      // get local ids of matching source points
      size_t * src_points = xmalloc(count * sizeof(*src_points));
      yac_intep_grid_src_global_to_local(
        interp_grid, 0, tgt_global_ids, count, src_points);
      free(tgt_global_ids);

      struct remote_points tgts = {
        .data =
          yac_interp_grid_get_tgt_remote_points(
            interp_grid, tgt_points, count),
        .count = count};

      struct remote_point * srcs =
        yac_interp_grid_get_src_remote_points(
          interp_grid, 0, src_points, count);

      yac_interp_weights_add_direct(weights, &tgts, srcs);

      free(tgts.data);
      free(tgt_points);
      free(srcs);
      free(src_points);
    }

    double *** src_data, ** tgt_data;

    if (is_src) {
      src_data = xmalloc(4 * sizeof(*src_data));
      for (size_t i = 0; i < 4; ++i)
        src_data[i] = xmalloc(1 * sizeof(**src_data));

      for (size_t collection_idx = 0; collection_idx < 4;
           ++collection_idx) {
        double * src_field =
          xmalloc(grid_data.num_cells * sizeof(*src_field));
        for (size_t i = 0; i < grid_data.num_cells; ++i)
          src_field[i] =
            (grid_data.core_cell_mask[i])?
              (double)(grid_data.cell_ids[i]):(-1.0);
        src_data[collection_idx][0] = src_field;
      }
      tgt_data = NULL;
    } else {
      tgt_data = xmalloc(4 * sizeof(*tgt_data));
      for (size_t collection_idx = 0; collection_idx < 4;
           ++collection_idx)
        tgt_data[collection_idx] =
          xmalloc(grid_data.num_cells * sizeof(**tgt_data));
      src_data = NULL;
    }

    // checking with different reorder types and collection sizes
    for (size_t reorder_type = 0; reorder_type < reorder_type_count;
         ++reorder_type)
      for (size_t collection_size = 1; collection_size <= 4;
           collection_size *= 2)
          if (check_direct_results(
                &grid_data, is_src, weights, reorder_types[reorder_type],
                collection_size, src_data, tgt_data))
            PUT_ERR("ERROR(yac_interp_weights_add_direct): "
                    "invalid interpolation result");

    // check writing and reading
    {
      char const * weight_file_name =
        "test_interp_weights_parallel_direct.nc";
      // write weights to file
      yac_interp_weights_write_to_file(
        weights, weight_file_name, "src_grid", "dst_grid", -1);

      // read weights
      struct interp_method * method_stack[2] = {
        yac_interp_method_file_new(
          weight_file_name, -1, "src_grid", "dst_grid"), NULL};
      struct interp_weights * weights_from_file =
        yac_interp_method_do_search(method_stack, interp_grid);
      yac_interp_method_delete(method_stack);

      // check results
      if (check_direct_results(
            &grid_data, is_src, weights, MAPPING_ON_SRC, 1, src_data, tgt_data))
        PUT_ERR("ERROR(yac_interp_weights_add_direct): "
                "invalid interpolation result (read from file)");
      if (comm_rank == 0) unlink(weight_file_name);
      yac_interp_weights_delete(weights_from_file);
    }

    if (is_src) {
      for (size_t collection_idx = 0; collection_idx < 4; ++collection_idx) {
        free(src_data[collection_idx][0]);
        free(src_data[collection_idx]);
      }
      free(src_data);
    } else {
      for (size_t collection_idx = 0; collection_idx < 4; ++collection_idx)
        free(tgt_data[collection_idx]);
      free(tgt_data);
    }

    yac_interp_weights_delete(weights);
    yac_interp_grid_delete(interp_grid);
    yac_dist_grid_pair_delete(grid_pair);
    yac_basic_grid_delete(grid);
    yac_mpi_call(MPI_Comm_free(&local_grid_comm), MPI_COMM_WORLD);
  }

  xt_finalize();

  MPI_Finalize();

  return TEST_EXIT_CODE;
}

static int check_direct_results(
  struct basic_grid_data * grid, int is_src, struct interp_weights * weights,
  enum interp_weights_reorder_type reorder_type, size_t collection_size,
  double *** src_data, double ** tgt_data) {

  int err_count = 0;

  struct interpolation * interpolation =
    yac_interp_weights_get_interpolation(
      weights, reorder_type, collection_size);

  if (!is_src)
    for (size_t collection_idx = 0; collection_idx < collection_size;
         ++collection_idx)
      for (size_t k = 0; k < grid->num_cells; ++k)
        tgt_data[collection_idx][k] = -1;

  yac_interpolation_execute(interpolation, src_data, tgt_data);

  if (!is_src) {
    for (size_t collection_idx = 0; collection_idx < collection_size;
         ++collection_idx) {
      for (size_t j = 0; j < grid->num_cells; ++j) {
        if (grid->core_cell_mask[j]) {
          if (tgt_data[collection_idx][j] !=
              (double)(grid->cell_ids[j])) err_count++;
        } else {
          if (tgt_data[collection_idx][j] != -1.0) err_count++;
        }
      }
    }
  }

  if (is_src) {
    yac_interpolation_execute_put(interpolation, src_data);
  } else {
    for (size_t collection_idx = 0; collection_idx < collection_size;
         ++collection_idx)
      for (size_t k = 0; k < grid->num_cells; ++k)
        tgt_data[collection_idx][k] = -1;
    yac_interpolation_execute_get(interpolation, tgt_data);
    for (size_t collection_idx = 0; collection_idx < collection_size;
         ++collection_idx) {
      for (size_t j = 0; j < grid->num_cells; ++j) {
        if (grid->core_cell_mask[j]) {
          if (tgt_data[collection_idx][j] !=
              (double)(grid->cell_ids[j])) err_count++;
        } else {
          if (tgt_data[collection_idx][j] != -1.0) err_count++;
        }
      }
    }
  }
  yac_interpolation_delete(interpolation);

  return err_count;
}

static void handle_error(int status) {
  if (status == NC_NOERR) return;
  const char *err_string;
  err_string = nc_strerror(status);
  printf ("%s \n", err_string);
  exit(EXIT_FAILURE);
}

static void get_basic_grid_data(
  char * filename, size_t * num_cells, size_t * num_vertices,
  size_t * num_edges) {

  int ncid;

  handle_error(nc_open(filename, NC_NOWRITE, &ncid));

  int dimid;
  handle_error(nc_inq_dimid(ncid, "cell", &dimid));
  handle_error(nc_inq_dimlen(ncid, dimid, num_cells));
  handle_error(nc_inq_dimid(ncid, "vertex", &dimid));
  handle_error(nc_inq_dimlen(ncid, dimid, num_vertices));
  handle_error(nc_inq_dimid(ncid, "edge", &dimid));
  handle_error(nc_inq_dimlen(ncid, dimid, num_edges));
}

static struct basic_grid_data
  generate_dummy_grid_data(
    size_t local_num_cells, size_t global_num_cells, size_t num_cells_offset) {

  double coordinates_x[global_num_cells + 1];
  double coordinates_y[2] = {0.0, 1.0};
  size_t num_cells[2] = {global_num_cells, 1};
  size_t local_start[2] = {num_cells_offset, 0};
  size_t local_count[2] = {local_num_cells, 1};

  for (size_t i = 0; i <= global_num_cells; ++i) coordinates_x[i] = (double)i;

  return
    yac_generate_basic_grid_data_reg2d(
      coordinates_x, coordinates_y, num_cells, local_start, local_count, 0);
}

static int check_results_ref(
  int is_src, struct interp_weights * weights,
  enum interp_weights_reorder_type reorder_type, double * src_data,
  double * tgt_data, double * ref_tgt_data, size_t tgt_size) {

  int err_count = 0;

  struct interpolation * interpolation =
    yac_interp_weights_get_interpolation(
      weights, reorder_type, 1);

  double * tgt_data_[1] = {tgt_data};
  double * src_data_[1] = {src_data};
  double ** src_data__[1] = {src_data_};

  if (!is_src)
    for (size_t k = 0; k < tgt_size; ++k) tgt_data[k] = -1;

  yac_interpolation_execute(interpolation, src_data__, tgt_data_);

  if (!is_src)
    for (size_t j = 0; j < tgt_size; ++j)
      if (tgt_data[j] != ref_tgt_data[j]) err_count++;

  if (is_src) {
    yac_interpolation_execute_put(interpolation, src_data__);
  } else {
    for (size_t k = 0; k < tgt_size; ++k) tgt_data[k] = -1;
    yac_interpolation_execute_get(interpolation, tgt_data_);
    for (size_t j = 0; j < tgt_size; ++j)
      if (tgt_data[j] != ref_tgt_data[j]) err_count++;
  }
  yac_interpolation_delete(interpolation);

  return err_count;
}

#else
int main(void) {return EXIT_SKIP_TEST;}
#endif // YAC_NETCDF_ENABLED
